local chunksize = 32

local regenerate_gui_name = "regenerate-terrain-frame"
local mirror_surface_index = "regnerate-terrain-mirror"

local function chunk_to_position(chunk, for_tiles)
  return {chunk.x * chunksize, chunk.y * chunksize}
end

local function chunk_to_area(chunk, for_tiles)
  if for_tiles then
    return {{chunk.x * chunksize, chunk.y * chunksize}, {chunk.x * chunksize + chunksize -1, chunk.y * chunksize + chunksize -1}}
  else
    return {{chunk.x * chunksize, chunk.y * chunksize}, {chunk.x * chunksize + chunksize-0.0001, chunk.y * chunksize + chunksize-0.0001}}
  end
end

local function regenerate_surface(target_surface)

  game.print({"regenerate-terrain.begin", target_surface.index})

  -- regenerate empty chunks
  -- for other chunks copy in tiles from a mirror surface
  -- if new tile is water place water and old tile was not then place landfill if near a player entity
  if game.surfaces[mirror_surface_index] then
      game.delete_surface(mirror_surface_index)
  end

  local process_legacy_chunk_tiles = true
  local process_natural_entities = true
  local process_respawned_chunk_entities = true

  -- mark chunks as "legacy" or "respawned"
  log("Rebuild Terrain: surface_index = " .. target_surface.index)

  local function log_chunk(chunk, message)
    --log("chunk "..chunk.x .." "..chunk.y.. " " .. message)
  end

  if not game.surfaces[mirror_surface_index] then
    game.create_surface(mirror_surface_index, target_surface.map_gen_settings)
  end
  local target_mirror_surface = game.surfaces[mirror_surface_index]

  local chunks = {}

  for chunk in target_surface.get_chunks() do
    -- geting the chunk creates it
    local chunkdata = chunk
    table.insert(chunks, chunkdata)
    chunkdata.created = game.tick

    chunkdata.protected = false

    if chunk.x == 0 and chunk.y == 0 then
      chunkdata.protected = true
    end
    -- find force entities from player forces
    if not chunkdata.protected then
      for _, force in pairs(game.forces) do
        if force.name ~= "neutral" and force.name ~= "enemy" then
          if target_surface.count_entities_filtered{
              area = chunk_to_area(chunk),
              force = force,
              limit = 1
            } > 0 then
              chunkdata.protected = true
              --log_chunk(chunk, "has entities of " .. force.name)
              break
          end
        end
      end
    end

    -- find rail
    if not chunkdata.protected then
      if target_surface.count_entities_filtered{
          area = chunk_to_area(chunk),
          type = "straight-rail",
          limit = 1
        } > 0 then
          chunkdata.protected = true
          --log_chunk(chunk, "has rail")
      end
    end

    if chunkdata.protected then

      -- there's stuff here so keep the chunk and replace per-tile
      chunkdata.type = "legacy"
      target_mirror_surface.request_to_generate_chunks({x = chunk.x * chunksize, y = chunk.y * chunksize}, 0)

    else

      --log_chunk(chunk, "can be respawned")
      -- delete the chunk then put any enemies back
      chunkdata.type = "respawned"
      chunkdata.pollution = target_surface.get_pollution({x = chunk.x * chunksize + chunksize / 2, y = chunk.y * chunksize + chunksize / 2})
      if chunkdata.pollution > 0 then
        log_chunk(chunk, "has pollution " .. chunkdata.pollution)
      end

      local enemy_entities = target_surface.find_entities_filtered{
        area = chunk_to_area(chunk),
        force = "enemy"
      }
      if #enemy_entities > 0 then
        --log_chunk(chunk, "has " .. #enemy_entities .." enemies")
        chunkdata.enemy = {}
        for _, entity in pairs(enemy_entities) do
          if entity.type ~= "unit" then -- don't bother with units for now
            table.insert(chunkdata.enemy, { type = entity.type, name = entity.name, position = entity.position })
          end
        end
      end

      -- no player forces, just delete and remake
      target_surface.delete_chunk(chunk)

    end
  end

  target_mirror_surface.force_generate_chunk_requests()


  local water_tiles = {
    ["water"] = true,
    ["deepwater"] = true,
    ["water-green"] = true,
    ["deepwater-green"] = true,
    ["water-shallow"] = true,
    ["water-mud"] = true
  }

  local floor_tiles = {}
  for _, tile in pairs(game.tile_prototypes) do
    if tile.mineable_properties.minable then
      floor_tiles[tile.name] = true
    end
  end

  -- set legacy tiles
  if process_legacy_chunk_tiles then


    for _, chunk in pairs(chunks) do
      if chunk.type == "legacy" then
        local tiles = {}
        --log_chunk(chunk, "legacy processing")
        for y = 0, 31, 1 do
          for x = 0, 31, 1 do
            local position = {
              x = chunk.x * chunksize + x,
              y = chunk.y * chunksize + y
            }
            local old_tile_name = target_surface.get_tile(position.x, position.y).name
            if not floor_tiles[old_tile_name] then
                local new_tile_name = target_mirror_surface.get_tile(position.x, position.y).name
                local new_is_water = water_tiles[new_tile_name]
                local old_is_water = water_tiles[old_tile_name]
                if new_is_water and not old_is_water then
                  local area = {{position.x, position.y}, {position.x+0.9999, position.y+0.9999}}
                  for _, force in pairs(game.forces) do
                    if force.name ~= "neutral" and force.name ~= "enemy" then
                      if target_surface.count_entities_filtered{
                          area = area,
                          force = force,
                          limit = 1
                        } > 0 then
                          new_tile_name = "landfill"
                          break
                      end
                    end
                  end
                  if target_surface.count_entities_filtered{
                      area = area,
                      type = "straight-rail",
                      limit = 1
                    } > 0 then
                      new_tile_name = "landfill"
                  end
                end
                table.insert(tiles, {name = new_tile_name, position = position})
            end
          end
        end
        local trees = target_surface.find_entities_filtered{type = "tree", area = chunk_to_area(chunk)}
        for _, tree in pairs(trees) do
          tree.destroy()
        end
        target_surface.set_tiles(tiles, true)
      end
    end
  end

  target_surface.regenerate_decorative() -- must be before deleted chunks generate
  --target_surface.regenerate_entity({"cliff", "small-cliff"})

  if process_natural_entities then
    local natural_entities = target_mirror_surface.find_entities()
    --log("natural_entities: " .. #natural_entities)
    for _, natural in pairs(natural_entities) do
        local valid  = natural.type ~= "resource" and natural.type~="cliff"
        local area = natural.selection_box or natural.bounding_box
        --log("Copying decorative " .. natural.name)
        if valid and floor_tiles[target_surface.get_tile(natural.position.x, natural.position.y).name] then
          valid = false
        end
        if (valid and (natural.minable or natural.destructible or natural.type=="cliff") and area) then
         --log("Minable, selection box " .. serpent.block(area))
         area.left_top.x = area.left_top.x - 2
         area.left_top.y = area.left_top.y - 2
         area.right_bottom.x = area.right_bottom.x + 1.999
         area.right_bottom.y = area.right_bottom.y + 1.999
          if (target_surface.count_entities_filtered{
            area = area,
            limit = 1,
            force = "player"} > 0)
           or (target_surface.count_entities_filtered{
            area = area,
            limit = 1,
            force = "enemy"} > 0) then
                valid = false
          end
        end
        if valid then
            local options = {
            name=natural.name,
              position=natural.position,
              direction=natural.direction,
              force=natural.force,
              graphics_variation=natural.graphics_variation,
              bounding_box=natural.bounding_box,
              secondary_bounding_box=natural.secondary_bounding_box,
              selection_box=natural.selection_box,
              secondary_selection_box=natural.secondary_selection_box
            }
            if options.type == "cliff" then
                cliff_orientation=natural.cliff_orientation
            end
           if (not target_surface.create_entity(options)) then log("Can't create entity") end
        end
    end
  end

  game.delete_surface(target_mirror_surface)

  for _, chunk in pairs(chunks) do
    -- warning: doing this during get chunks adds it to the current loop
    target_surface.request_to_generate_chunks({x = chunk.x * chunksize, y = chunk.y * chunksize}, 0)
  end

  target_surface.force_generate_chunk_requests()

  if process_respawned_chunk_entities then
    for _, chunk in pairs(chunks) do
      if chunk.type == "respawned" then

        --log_chunk(chunk, "respawned processing")
        if chunk.enemy then
          for _, entity in pairs(chunk.enemy) do
            if target_surface.can_place_entity(entity) then
              target_surface.create_entity(entity)
            end
          end
        end
        if chunk.pollution and chunk.pollution > 0 then
          --log_chunk(chunk, "apply pollution " .. chunk.pollution)
          --target_surface.pollute({x = chunk.x * chunksize + chunksize / 2, y = chunk.y * chunksize + chunksize / 2}, chunk.pollution)
          -- poluting an newly respaned chunk does not work anymore.
          target_surface.pollute({0,0}, chunk.pollution)
        end
      end
    end
  end

  game.print({"regenerate-terrain.complete", target_surface.index})
end


local function build_regenerate_gui(player)
  if player.gui.center[regenerate_gui_name] then
    player.gui.center[regenerate_gui_name].destroy()
  end

  local frame = player.gui.center.add{
    type = "frame", name = regenerate_gui_name, caption = {"regenerate-terrain.frame-heading"}, direction = "vertical", style = "regenerate-terrain_frame"}
  frame.style.width = 400
  frame.style.maximal_height = 600

  local warning1 = frame.add{ type = "label", name = "warning-detail-1", caption = {"regenerate-terrain.warning-detail-1"},  style = "regenerate-terrain_label_multiline"}
  local warning2 = frame.add{ type = "label", name = "warning-detail-2", caption = {"regenerate-terrain.warning-detail-2"},  style = "regenerate-terrain_label_multiline"}
  local warning3 = frame.add{ type = "label", name = "warning-detail-3", caption = {"regenerate-terrain.warning-detail-3"},  style = "regenerate-terrain_label_multiline"}
  warning2.style.top_margin = 5
  warning3.style.top_margin = 5

  local surfaces_list = frame.add{ type = "scroll-pane", name = "surfaces_list", direction = "vertical", style = "scroll_pane_with_dark_background_under_subheader" }
  --local surfaces_list = frame.add{ type = "scroll-pane", name = "surfaces_list", direction = "vertical" }
  surfaces_list.style.maximal_height = 300
  surfaces_list.style.width = 380
  surfaces_list.style.top_margin = 10
  surfaces_list.style.bottom_margin = 10
  surfaces_list.vertical_scroll_policy = "always"

  for index, surface in pairs(game.surfaces) do
    if index ~= mirror_surface_index then
      surfaces_list.add{ type = "radiobutton", style = "regenerate-terrain_radiobutton", name = index, caption = index .. (surface.index == player.surface.index and " (current)" or ""), state = surface.index == player.surface.index }
    end
  end

  local warning4 = frame.add{ type = "label", name = "warning-detail-4", caption = {"regenerate-terrain.warning-detail-4"},  style = "regenerate-terrain_label_multiline"}
  warning4.style.top_margin = 5

  local buttoncontainer = frame.add{ type = "flow", name = "buttonflow", direction = "horizontal"}

  buttoncontainer.add{ type = "button", name = "regenerate-terrain-cancel", style = "back_button", caption = {"regenerate-terrain.cancel"} }
  buttoncontainer.style.minimal_width = 380
  buttoncontainer.style.horizontally_stretchable = true

  local confirm = buttoncontainer.add{ type = "button", name = "regenerate-terrain-confirm", style = "red_confirm_button", caption = {"regenerate-terrain.confirm"} }
  confirm.style.horizontal_align = "right"
  confirm.style.left_margin = 10

end

local function close_regenerate_gui(player)
  if player.gui.center[regenerate_gui_name] then
    player.gui.center[regenerate_gui_name].destroy()
  end
end

local function toggle_regenerate_gui(player)
  if player.gui.center[regenerate_gui_name] then
    close_regenerate_gui(player)
  else
    build_regenerate_gui(player)
  end
end

local function on_gui_checked_state_changed(event)
  if event.player_index
    and game.players[event.player_index]
    and game.players[event.player_index].connected
    and game.players[event.player_index].valid then
      local player = game.players[event.player_index]
      if event.element and event.element.valid then
        local element = event.element
        if element.parent.name == "surfaces_list" then
          for _, child in pairs(element.parent.children) do
            child.state = false
          end
          element.state = true
        end
      end
  end
end

local function on_gui_click(event)
  if event.player_index
    and game.players[event.player_index]
    and game.players[event.player_index].connected
    and game.players[event.player_index].valid then
      local player = game.players[event.player_index]
      if event.element and event.element.valid then
        local element = event.element
        if element.name == "regenerate-terrain-cancel" then
          close_regenerate_gui(player)
          return
        end
        if element.name == "regenerate-terrain-confirm" then
          local target_surface_index = nil
          if player.gui.center[regenerate_gui_name]
            and player.gui.center[regenerate_gui_name].surfaces_list then
              for _, child in pairs(player.gui.center[regenerate_gui_name].surfaces_list.children) do
                if child.state then
                  target_surface_index = child.name
                end
              end
          end
          if target_surface_index and game.get_surface(target_surface_index) then
            local surface = game.get_surface(target_surface_index)
            regenerate_surface(surface)
            close_regenerate_gui(player)
          else
            game.print({"regenerate-terrain.no-surface", target_surface_index})
          end
          return
        end
      end
  end
end

local function on_lua_shortcut(event)
  if event.prototype_name == "regenerate-terrain"
    and event.player_index
    and game.players[event.player_index]
    and game.players[event.player_index].connected
    then
      toggle_regenerate_gui(game.players[event.player_index])
  end
end

script.on_event(defines.events.on_gui_click, on_gui_click)
script.on_event(defines.events.on_gui_checked_state_changed, on_gui_checked_state_changed)
script.on_event(defines.events.on_lua_shortcut, on_lua_shortcut)
